#ifndef lint
static char RCSid[] = "ttcp.c $Revision: 1.12 $";
#endif

#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>		/* struct timeval */
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/mman.h>

#include <sys/resource.h>

struct sockaddr_in sinme;
struct sockaddr_in sinhim;
struct sockaddr_in frominet;

int nfd;                        /* fd for child processes... */
int serv_child =0;
int childpid;
int num_client = 1;
int *client_cld_pid;
int d;
int parent_gid;
int parentsig = 0;
int childsig = 0;
int iteration = 1;
int tries = 0;
long double *it_p;

static struct   timeval time0;  /* Time at which timing started */
static struct   rusage ru0;     /* Resource utilization at the start */
struct timeval timedol;

double first_time;
double last_time;
double total_bytes;

typedef struct {
  int pid;
  double s_t;
  double e_t;
  int buflen;
  int nbuf;
  int bufoffset;
  int bufalign;
  int udp;
  int options;
  int one;
  short port;
  int nodelay;
  int sockbufsize;
  double nbytes;
  unsigned long numCalls;
  double cput;
  double realt;
} CHILD_DATA_STRUCT;

CHILD_DATA_STRUCT *cdstp;


/*
 * Shared Memory variables.
 */

key_t shmkey;
int shmflg;
int shmid;
int shmsize;
char *shm;
int *shma;

/*
 * Semaphore variables.
 */

struct msemaphore *Sem;

int domain, fromlen;
int fd;				/* fd of network socket */

int buflen = 8 * 1024;		/* length of buffer */
char *buf;			/* ptr to dynamic buffer */
int nbuf = 2 * 1024;		/* number of buffers to send in sinkmode */

int bufoffset = 0;		/* align buffer to this */
int bufalign = 16*1024;		/* modulo this */

int udp = 0;			/* 0 = tcp, !0 = udp */
int options = 0;		/* socket options */
int one = 1;                    /* for 4.3 BSD style setsockopt() */
short port = 5001;		/* TCP port number */
char *host;			/* ptr to name of host */
int trans;			/* 0=receive, !0=transmit mode */
int sinkmode = 1;		/* 0=normal I/O, !0=sink/source mode */
int verbose = 0;		/* 0=print basic info, 1=print cpu rate, proc
				 * resource usage. */
int nodelay = 0;		/* set TCP_NODELAY socket option */
int b_flag = 0;			/* use mread() */
int sockbufsize = 0;		/* socket buffer size to use */
char fmt = 'K';			/* output format: k = kilobits, K = kilobytes,
				 *  m = megabits, M = megabytes, 
				 *  g = gigabits, G = gigabytes */
int touchdata = 0;		/* access data after reading */

struct hostent *addr;
extern int errno;
extern int optind;
extern char *optarg;

char Usage[] = "\
Usage: ttcp -t [-options] host [ < in ]\n\
       ttcp -r [-options > out]\n\
Common options:\n\
	-l ##	length of bufs read from or written to network (default 8192)\n\
	-u	use UDP instead of TCP\n\
	-p ##	port number to send to or listen at (default 5001)\n\
	-s	-t: source a pattern to network\n\
		-r: sink (discard) all data from network\n\
	-A	align the start of buffers to this modulus (default 16384)\n\
	-O	start buffers at this offset from the modulus (default 0)\n\
	-v	verbose: print more statistics\n\
	-d	set SO_DEBUG socket option\n\
	-b ##	set socket buffer size (if supported)\n\
	-f X	format for rate: k,K = kilo{bit,byte}; m,M = mega; g,G = giga\n\
Options specific to -t:\n\
	-n##	number of source bufs written to network (default 2048)\n\
	-D	don't buffer TCP writes (sets TCP_NODELAY socket option)\n\
Options specific to -r:\n\
	-B	for -s, only output full blocks as specified by -l (for TAR)\n\
	-T	\"touch\": access each byte as it's read\n\
";	

char stats[128];
double nbytes;			/* bytes on net */
unsigned long numCalls;		/* # of I/O system calls */
double cput, realt;		/* user, real time (seconds) */

CHILD_DATA_STRUCT *cdsta();
void err();
void mes();
int pattern();
void prep_timer();
double read_timer();
int Nread();
int Nwrite();
void delay();
int mread();
char *outfmt();
void semup ();
void semdown ();

void
sigpipe()
{
}

void SigToParent() {
  parentsig = 1;
}

void SigToChild() {
  childsig = 1;
}  

main(argc,argv)
    int argc;
    char **argv;
{
  unsigned long addr_tmp;
  int c;
  
  if (argc < 2) goto usage;
  
  while ((c = getopt(argc, argv, "drstuvBDTb:f:l:n:p:A:O:c:i:")) != -1) {
    switch (c) {
      
    case 'B':
      b_flag = 1;
      break;
    case 't':
      trans = 1;
      break;
    case 'r':
      trans = 0;
      break;
    case 'd':
      options |= SO_DEBUG;
      break;
    case 'D':
#ifdef TCP_NODELAY
      nodelay = 1;
#else
      fprintf(stderr, 
	      "ttcp: -D option ignored: TCP_NODELAY socket option not supported\n");
#endif
      break;
      
    case 'c':
      num_client = atoi(optarg);
      break;
    case 'n':
      nbuf = atoi(optarg);
      break;
    case 'l':
      buflen = atoi(optarg);
      break;
    case 's':
      sinkmode = !sinkmode;
      break;
    case 'p':
      port = atoi(optarg);
      break;
    case 'u':
      udp = 1;
      break;
    case 'v':
      verbose = 1;
      break;
    case 'i':
      iteration = atoi(optarg);
    case 'A':
      bufalign = atoi(optarg);
      break;
    case 'O':
      bufoffset = atoi(optarg);
      break;
    case 'b':
#if defined(SO_SNDBUF) || defined(SO_RCVBUF)
      sockbufsize = atoi(optarg);
#else
      fprintf(stderr, 
	      "ttcp: -b option ignored: SO_SNDBUF/SO_RCVBUF socket options not supported\n");
#endif
      break;
    case 'f':
      fmt = *optarg;
      break;
    case 'T':
      touchdata = 1;
      break;
      
    default:
      goto usage;
    }
  }
  if(trans) {
    if ((it_p = (long double *)malloc(sizeof(long double)*iteration)) == NULL) {
      perror("Can't allocate for itteration... ");
      exit(20);
    }
  iterate:
    tries++;
    parentsig=0;
    /*
     * Shared Memory Setup
     */
    shmkey = 1995;
    shmsize = sizeof(int)*2 + sizeof(CHILD_DATA_STRUCT)*num_client + sizeof(msemaphore);
    shmflg = 0666 | IPC_CREAT;
    if ((shmid = shmget (shmkey, shmsize, shmflg)) == -1) {
      perror ("shmget FAILED!!!");
      exit(10);
    }
    shm = (char *) shmat(shmid, 0, 0);
    shma = (int *)shm;

    /*
     * Semaphore Setup
     */
    Sem = (msemaphore *)(shm+sizeof(int)*2 + sizeof(CHILD_DATA_STRUCT)*num_client);
    if((msem_init(Sem, 0)) == NULL) perror("MSEM_INIT ERROR!!!");

    (*shma) = num_client;
    client_cld_pid = (int *)malloc(num_client*sizeof(int));
    signal(SIGUSR1, SigToParent);
    signal(SIGUSR2, SigToChild);
    sigblock (sigmask(SIGUSR1));
    for (d=0;d<num_client;d++) {
      if ((*(client_cld_pid+sizeof(int)*d) = fork()) == 0) goto after_clients_loop;
    }
    /*
     * ONLY THE PARENT
     */
    if ((setpgrp (0,(parent_gid = getpid()))) < 0) perror("Cant set my own group id!!!");
    for (d=0;d<num_client;d++)
	if ((setpgrp((*(client_cld_pid+sizeof(int)*d)), parent_gid)) < 0)
	    perror("Can't set child group id!!!");
    while (parentsig == 0) sigpause(0);
    parentsig = 0;
    signal(SIGUSR1, SigToParent);
    sigblock (sigmask(SIGUSR1));
    /* printf("All children have connected!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"); */
    signal(SIGUSR2, SIG_IGN);  /* ignore the broadcasted signal */
    kill(0,SIGUSR2);
    while (parentsig == 0) sigpause(0);
    /* printf("All children have Finished!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"); */
    cdstp = cdsta(0);
    first_time = cdstp->s_t;
    last_time = cdstp->e_t;
    total_bytes = cdstp->nbytes;
    for (d=1;d<num_client;d++) {
      cdstp = cdsta(d);
      if (cdstp->s_t < first_time) first_time = cdstp->s_t;
      if (cdstp->e_t > last_time) last_time = cdstp->e_t;
      total_bytes+=cdstp->nbytes;
      /* printf("Child %d:PID:\t%d\tST:\t%.2f\tET:\t%.2f\n",d,cdstp->pid,cdstp->s_t,cdstp->e_t); */
    }
    *(it_p+(tries-1)*sizeof(long double)) = (long double)(total_bytes/(last_time-first_time));
    printf("The bandwith for try %d: %f B/sec.\n",tries,(total_bytes/(last_time-first_time)));

    
    /*
     * Parent Cleanup.
     */
    /* if ((semctl(semid, 1, IPC_RMID, 0)) < 0) perror("semctl: Can't remove semaphore\n"); */
    free(client_cld_pid);
    if (shmdt(shm) < 0)
	perror("Client CAN NOT detach shared memory.\n");
    if (shmctl(shmid, IPC_RMID, (struct shmid_ds *) 0) < 0)
	perror("Client CAN NOT remove shared memory.\n");
    if (iteration > tries) {
      sleep(10);
      goto iterate;
    }
    for (d=1;d<iteration; d++) {
      *it_p+=*(it_p+sizeof(long double)*d);
    }
    printf("The mean of iteration [%d] : %f B/sec.\n",iteration,((*it_p)/iteration));
    exit(0);
  }
 after_clients_loop:
  if(trans)  {
    
    if (optind == argc) goto usage;
    bzero((char *)&sinhim, sizeof(sinhim));
    host = argv[optind];
    if (atoi(host) > 0 )  {
      /* Numeric */
      sinhim.sin_family = AF_INET;
#if defined(cray)
      addr_tmp = inet_addr(host);
      sinhim.sin_addr = addr_tmp;
#else
      sinhim.sin_addr.s_addr = inet_addr(host);
#endif
    }
    else {
      if ((addr=gethostbyname(host)) == NULL)
	err("bad hostname");
      sinhim.sin_family = addr->h_addrtype;
      bcopy(addr->h_addr,(char*)&addr_tmp, addr->h_length);
#if defined(cray)
      sinhim.sin_addr = addr_tmp;
#else
      sinhim.sin_addr.s_addr = addr_tmp;
#endif /* cray */
    }
    sinhim.sin_port = htons(port);
    sinme.sin_port = 0;		/* free choice */
  }
  else {
    /* rcvr */
    sinme.sin_port =  htons(port);
  }
  
  
  if (udp && buflen < 5) {
    buflen = 5;		/* send more than the sentinel size */
  }
  
  if ( (buf = (char *)malloc(buflen+bufalign)) == (char *)NULL)
    err("malloc");
  if (bufalign != 0)
    buf +=(bufalign - ((int)buf % bufalign) + bufoffset) % bufalign;
  
  if ((fd = socket(AF_INET, udp?SOCK_DGRAM:SOCK_STREAM, 0)) < 0)
    err("socket");
  if (bind(fd, &sinme, sizeof(sinme)) < 0)
    err("bind");
  
#if defined(SO_SNDBUF) || defined(SO_RCVBUF)
  if (sockbufsize) {
    if (trans) {
      if (setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &sockbufsize,
		     sizeof sockbufsize) < 0)
	err("setsockopt: sndbuf");
    }
    else {
      if (setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &sockbufsize,
		     sizeof sockbufsize) < 0)
	err("setsockopt: rcvbuf");
    }
  }
#endif
  
  if (!udp)  {
    signal(SIGPIPE, sigpipe);
    if (trans) {
      /* We are the client if transmitting */
      if (options)  {
	if( setsockopt(fd, SOL_SOCKET, options, &one, sizeof(one)) < 0)
	  err("setsockopt");
      }
#ifdef TCP_NODELAY
      if (nodelay) {
	struct protoent *p;
	p = getprotobyname("tcp");
	if( p && setsockopt(fd, p->p_proto, TCP_NODELAY, 
			    &one, sizeof(one)) < 0)
	  err("setsockopt: nodelay");
      }
#endif
      if(connect(fd, &sinhim, sizeof(sinhim) ) < 0)
	err("connect");
      msem_lock(Sem, 0);
      (*shma)--;
      if ((*shma) == 0) kill(getppid(),SIGUSR1);
      msem_unlock(Sem, 0);
      while (childsig == 0) sigpause(0);
    }
    else {
      /* otherwise, we are the server and 
       * should listen for the connections
       */
#if defined(ultrix) || defined(sgi)
      listen(fd,1);   /* workaround for alleged u4.2 bug */
#else
      listen(fd,0);   /* allow a queue of 0 */
#endif
      if(options)  {
	if( setsockopt(fd, SOL_SOCKET, options, &one, sizeof(one)) < 0)
	  err("setsockopt");
      }
      fromlen = sizeof(frominet);
      domain = AF_INET;
      for (;;) {
	if((nfd=accept(fd, &frominet, &fromlen) ) < 0)
	  err("accept");
	{ struct sockaddr_in peer;
	  int peerlen = sizeof(peer);
	  if (getpeername(nfd, (struct sockaddr_in *) &peer, 
			  &peerlen) < 0) {
	    err("getpeername");
	  }
	  /* fprintf(stderr,"ttcp-r: accept from %s\n",inet_ntoa(peer.sin_addr)); */
	  serv_child++;
	  printf("%d ",serv_child);
	  fflush(stdout);
	}
	if ((childpid = fork()) > 0) close(nfd);
	else if (childpid == 0) goto afterloop;
      }
    afterloop:
      close(fd);
    }
  }
  errno = 0;
  if (sinkmode) {      
    register int cnt;
    if (trans)  {
      pattern( buf, buflen );
      prep_timer();
      if(udp)  (void)Nwrite( fd, buf, 4 ); /* rcvr start */
      while (nbuf-- && Nwrite(fd,buf,buflen) == buflen)
	nbytes += buflen;
      if(udp)  (void)Nwrite( fd, buf, 4 ); /* rcvr end */
    }
    else {
      if (udp) {
	while ((cnt=Nread(nfd,buf,buflen)) > 0)  {
	  static int going = 0;
	  if( cnt <= 4 )  {
	    if( going )
	      break;	/* "EOF" */
	    going = 1;
	    prep_timer();
	  }
	  else {
	    nbytes += cnt;
	  }
	}
      }
      else {
	while ((cnt=Nread(nfd,buf,buflen)) > 0)  {
	  nbytes += cnt;
	}
      }
    }
  }
  else {
    register int cnt;
    if (trans)  {
      while((cnt=read(0,buf,buflen)) > 0 &&
	    Nwrite(fd,buf,cnt) == cnt)
	nbytes += cnt;
    }
    else  {
      while((cnt=Nread(nfd,buf,buflen)) > 0 &&
	    write(1,buf,cnt) == cnt)
	nbytes += cnt;
    }
  }
  if(errno) err("IO");
  (void)read_timer(stats,sizeof(stats));
  if(udp&&trans)  {
    (void)Nwrite( fd, buf, 4 ); /* rcvr end */
    (void)Nwrite( fd, buf, 4 ); /* rcvr end */
    (void)Nwrite( fd, buf, 4 ); /* rcvr end */
    (void)Nwrite( fd, buf, 4 ); /* rcvr end */
  }
  if( cput <= 0.0 )  cput = 0.001;
  if( realt <= 0.0 )  realt = 0.001;
  /*
  fprintf(stdout,
	  "ttcp%s: Pid %d: %.0f bytes in %.2f real seconds = %s/sec +++\n",
	  trans?"-t":"-r", getpid(),
	  nbytes, realt, outfmt(nbytes/realt));
  if (verbose) {
    fprintf(stdout,
	    "ttcp%s: %.0f bytes in %.2f CPU seconds = %s/cpu sec\n",
	    trans?"-t":"-r",
	    nbytes, cput, outfmt(nbytes/cput));
  }
  fprintf(stdout,
	  "ttcp%s: %d I/O calls, msec/call = %.2f, calls/sec = %.2f\n",
	  trans?"-t":"-r",
	  numCalls,
	  1024.0 * realt/((double)numCalls),
	  ((double)numCalls)/realt);
  fprintf(stdout,"ttcp%s: %s\n", trans?"-t":"-r", stats);
  if (verbose) {
    fprintf(stdout,
	    "ttcp%s: buffer address %#x\n",
	    trans?"-t":"-r",
	    buf);
  }
  */
  if (trans) {
    msem_lock(Sem, 0);
    cdstp = cdsta((*shma));

    cdstp->pid = getpid();
    cdstp->s_t = time0.tv_sec + ((double)time0.tv_usec)/1000000;
    cdstp->e_t = timedol.tv_sec + ((double)timedol.tv_usec)/1000000;
    cdstp->buflen = buflen;
    cdstp->nbuf = nbuf;
    cdstp->bufoffset = bufoffset;
    cdstp->bufalign = bufalign;
    cdstp->udp = udp;
    cdstp->options = options;
    cdstp->one = one;
    cdstp->port = port;
    cdstp->nodelay = nodelay;
    cdstp->sockbufsize = sockbufsize;
    cdstp->nbytes = nbytes;
    cdstp->numCalls = numCalls;
    cdstp->cput = cput;
    cdstp->realt = realt;

    (*shma)++;

    if ((*shma) == num_client) {
      sleep(2);
      /* printf("SEnding Kill to the parent\n"); */
      kill(getppid(), SIGUSR1);
      /* printf("Last Child Finished!\n"); */
    }
    msem_unlock(Sem, 0);
  }
  exit(0);
  
 usage:
  fprintf(stderr,Usage);
  exit(1);
}

void
err(s)
     char *s;
{
  fprintf(stderr,"ttcp%s: ", trans?"-t":"-r");
  perror(s);
  fprintf(stderr,"errno=%d\n",errno);
  exit(1);
}

void
mes(s)
     char *s;
{
  fprintf(stderr,"ttcp%s: %s\n", trans?"-t":"-r", s);
}

pattern( cp, cnt )
     register char *cp;
     register int cnt;
{
  register char c;
  c = 0;
	while( cnt-- > 0 )  {
	  while( !isprint((c&0x7F)) )  c++;
	  *cp++ = (c++&0x7F);
	}
}

char *
outfmt(b)
     double b;
{
  static char obuf[50];
  switch (fmt) {
  case 'G':
    sprintf(obuf, "%.2f GB", b / 1024.0 / 1024.0 / 1024.0);
    break;
  default:
  case 'K':
    sprintf(obuf, "%.2f KB", b / 1024.0);
    break;
  case 'M':
    sprintf(obuf, "%.2f MB", b / 1024.0 / 1024.0);
    break;
  case 'g':
    sprintf(obuf, "%.2f Gbit", b * 8.0 / 1024.0 / 1024.0 / 1024.0);
    break;
  case 'k':
    sprintf(obuf, "%.2f Kbit", b * 8.0 / 1024.0);
    break;
  case 'm':
    sprintf(obuf, "%.2f Mbit", b * 8.0 / 1024.0 / 1024.0);
    break;
  }
  return obuf;
}

static void prusage();
static void tvadd();
static void tvsub();
static void psecs();


/*
 *			P R E P _ T I M E R
 */
void
prep_timer()
{
  gettimeofday(&time0, (struct timezone *)0);
  getrusage(RUSAGE_SELF, &ru0);
}

/*
 *			R E A D _ T I M E R
 * 
 */
double
read_timer(str,len)
     char *str;
{
  struct rusage ru1;
  struct timeval td;
  struct timeval tend, tstart;
  char line[132];
  
  getrusage(RUSAGE_SELF, &ru1);
  gettimeofday(&timedol, (struct timezone *)0);
  prusage(&ru0, &ru1, &timedol, &time0, line);
  (void)strncpy( str, line, len );
  
  /* Get real time */
  tvsub( &td, &timedol, &time0 );
  realt = td.tv_sec + ((double)td.tv_usec) / 1000000;
  
  /* Get CPU time (user+sys) */
  tvadd( &tend, &ru1.ru_utime, &ru1.ru_stime );
  tvadd( &tstart, &ru0.ru_utime, &ru0.ru_stime );
  tvsub( &td, &tend, &tstart );
  cput = td.tv_sec + ((double)td.tv_usec) / 1000000;
  if( cput < 0.00001 )  cput = 0.00001;
  return( cput );
}

static void
prusage(r0, r1, e, b, outp)
     register struct rusage *r0, *r1;
     struct timeval *e, *b;
     char *outp;
{
  struct timeval tdiff;
  register time_t t;
  register char *cp;
  register int i;
  int ms;
  
  t = (r1->ru_utime.tv_sec-r0->ru_utime.tv_sec)*100+
    (r1->ru_utime.tv_usec-r0->ru_utime.tv_usec)/10000+
      (r1->ru_stime.tv_sec-r0->ru_stime.tv_sec)*100+
	(r1->ru_stime.tv_usec-r0->ru_stime.tv_usec)/10000;
  ms =  (e->tv_sec-b->tv_sec)*100 + (e->tv_usec-b->tv_usec)/10000;
  
#define END(x)	{while(*x) x++;}
  cp = "%Uuser %Ssys %Ereal %P %Xi+%Dd %Mmaxrss %F+%Rpf %Ccsw";
  for (; *cp; cp++)  {
    if (*cp != '%')
      *outp++ = *cp;
    else if (cp[1]) switch(*++cp) {
      
    case 'U':
      tvsub(&tdiff, &r1->ru_utime, &r0->ru_utime);
      sprintf(outp,"%d.%01d", tdiff.tv_sec, tdiff.tv_usec/100000);
      END(outp);
      break;
      
    case 'S':
      tvsub(&tdiff, &r1->ru_stime, &r0->ru_stime);
      sprintf(outp,"%d.%01d", tdiff.tv_sec, tdiff.tv_usec/100000);
      END(outp);
      break;
      
    case 'E':
      psecs(ms / 100, outp);
      END(outp);
      break;
      
    case 'P':
      sprintf(outp,"%d%%", (int) (t*100 / ((ms ? ms : 1))));
      END(outp);
      break;
      
    }
  }
  *outp = '\0';
}

static void
tvadd(tsum, t0, t1)
     struct timeval *tsum, *t0, *t1;
{

  tsum->tv_sec = t0->tv_sec + t1->tv_sec;
  tsum->tv_usec = t0->tv_usec + t1->tv_usec;
  if (tsum->tv_usec > 1000000)
    tsum->tv_sec++, tsum->tv_usec -= 1000000;
}

static void
tvsub(tdiff, t1, t0)
     struct timeval *tdiff, *t1, *t0;
{
  
  tdiff->tv_sec = t1->tv_sec - t0->tv_sec;
  tdiff->tv_usec = t1->tv_usec - t0->tv_usec;
  if (tdiff->tv_usec < 0)
    tdiff->tv_sec--, tdiff->tv_usec += 1000000;
}

static void
psecs(l,cp)
     long l;
     register char *cp;
{
  register int i;
  
  i = l / 3600;
  if (i) {
    sprintf(cp,"%d:", i);
    END(cp);
    i = l % 3600;
    sprintf(cp,"%d%d", (i/60) / 10, (i/60) % 10);
    END(cp);
  }
  else {
    i = l;
    sprintf(cp,"%d", i / 60);
    END(cp);
  }
  i %= 60;
  *cp++ = ':';
  sprintf(cp,"%d%d", i / 10, i % 10);
}

/*
 *			N R E A D
 */
Nread( fd, buf, count )
     int fd;
     void *buf;
     int count;
{
  struct sockaddr_in from;
  int len = sizeof(from);
  register int cnt;
  if( udp )  {
    cnt = recvfrom( fd, buf, count, 0, &from, &len );
    numCalls++;
  }
  else {
    if( b_flag )
      cnt = mread( fd, buf, count );	/* fill buf */
    else {
      cnt = read( fd, buf, count );
      numCalls++;
    }
    if (touchdata && cnt > 0) {
      register int c = cnt, sum;
      register char *b = buf;
      while (c--)
	sum += *b++;
    }
  }
  return(cnt);
}

/*
 *			N W R I T E
 */
Nwrite( fd, buf, count )
     int fd;
     void *buf;
     int count;
{
  register int cnt;
  if( udp )  {
  again:
    cnt = sendto( fd, buf, count, 0, &sinhim, sizeof(sinhim) );
    numCalls++;
    if( cnt<0 && errno == ENOBUFS )  {
      delay(18000);
      errno = 0;
      goto again;
    }
  }
  else {
    cnt = write( fd, buf, count );
    numCalls++;
  }
  return(cnt);
}

void
delay(us)
{
  struct timeval tv;
  
  tv.tv_sec = 0;
  tv.tv_usec = us;
  (void)select( 1, (char *)0, (char *)0, (char *)0, &tv );
}

/*
 *			M R E A D
 *
 * This function performs the function of a read(II) but will
 * call read(II) multiple times in order to get the requested
 * number of characters.  This can be necessary because
 * network connections don't deliver data with the same
 * grouping as it is written with.  Written by Robert S. Miles, BRL.
 */
int
mread(fd, bufp, n)
    int fd;
    register char	*bufp;
    unsigned	n;
{
  register unsigned	count = 0;
  register int		nread;
  
  do {
    nread = read(fd, bufp, n-count);
    numCalls++;
    if(nread < 0)  {
      perror("ttcp_mread");
      return(-1);
    }
    if(nread == 0)
      return((int)count);
    count += (unsigned)nread;
    bufp += nread;
  } while(count < n);
  
  return((int)count);
}


CHILD_DATA_STRUCT *cdsta(index)
    int index;
{
  return (shm+sizeof(int)*2 + sizeof(CHILD_DATA_STRUCT)*index);
}

